/* pdbtool.c
   Program for testing the library in palmpack.a
   (also provides illustrative examples)
   Copyright (C) 1999-12-14 Dan Hovang (compression)
   Copyright (C) 1999-12-14 Linus Walleij (file formats)

This file is part of the PDBlib package.
Latest version: http://www.df.lth.se/~triad/krad/
Mail us: triad@df.lth.se

PDBlib is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with PDBlib; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "config.h"
#if HAVE_STDBOOL_H
# include <stdbool.h>
#else
typedef enum {false = 0, true = 1} bool;
#endif
#include <stdio.h>
#include "pilotpdb.h"

#define STR_MAXLEN      255

void DumpRecords(tPilotRecord *listp, FILE *fout)
{
  tPilotRecord *iterator;
  register DWORD i, l;
  register int j, k;

  iterator = listp;

  /* Dump each record */
  j = 0;
  while(iterator != NULL) {
    fprintf(fout, "Record: %d\n", j);
    k = 0;
    for (i = 0; i < iterator->dwChunkSize; i ++) {
      if (k == 0)
	fprintf(fout, "%08lX ", i);
      if (k == 8) {
	fprintf(fout, " ");
      }
      if (k == 16) {
	fprintf(fout, "  ");
	for (l = i - 16; l < i; l ++) {
	  fprintf(fout, "%c", (iterator->chunk[l] < 0x20 || iterator->chunk[l] > 0x7E) 
		  ? '.' : (char) iterator->chunk[l]);
	}
	fprintf(fout, "\n");
	if (i != iterator->dwChunkSize)
	  fprintf(fout, "%08lX ", i);
	k = 0;
      }
      fprintf(fout, " %02X", (unsigned char) iterator->chunk[i]);
      k ++;
    }
    if (k) {
      if (k == 16)
	fprintf(fout, "  ");
      else for (l = k; l < 17; l ++)
	fprintf(fout, "   ");
      for (l = i - k; l < i; l ++) {
	fprintf(fout, "%c", (iterator->chunk[l] < 0x20 || iterator->chunk[l] > 0x7E)
		? '.' : (char) iterator->chunk[l]);
      }
    }
    /* DocCompress(iterator->chunk, (int) iterator->dwChunkSize); */
    fprintf(fout, "\n");
    iterator = iterator->next;
    j ++;
  }
}

static DWORD txtfread(byte *buf, DWORD BufSize, FILE *fin)
{
  static byte cmpbuf[4];
  static byte cmpbufSZ = 0;
  register DWORD returnSZ;

  returnSZ = 0;
  while (!feof(fin) && returnSZ < BufSize) {
    if (cmpbufSZ < 4)
      cmpbufSZ += fread(cmpbuf + cmpbufSZ, 1, 4 - cmpbufSZ, fin);
    if (cmpbufSZ < 4) {
      memcpy(buf+returnSZ, cmpbuf, cmpbufSZ);
      returnSZ += cmpbufSZ;
      return (returnSZ);
    }
    /* Match double-byte linefeeds */
    if ((cmpbuf[0] == (unsigned char) 0x0D &&
	 cmpbuf[1] == (unsigned char) 0x0A &&
	 cmpbuf[2] == (unsigned char) 0x0D &&
	 cmpbuf[3] == (unsigned char) 0x0A) ||
	(cmpbuf[0] == (unsigned char) 0x0A &&
	 cmpbuf[1] == (unsigned char) 0x0D &&
	 cmpbuf[2] == (unsigned char) 0x0A &&
	 cmpbuf[3] == (unsigned char) 0x0D)) {
      /* Cut off 2 bytes replace with 0x0A 0x0A */
      buf[returnSZ] = 0x0A;
      returnSZ ++;
      buf[returnSZ] = 0x0A;
      returnSZ ++;
      cmpbufSZ = 0;
    }
    /* Match MacIntosh linefeeds */
    else if ((cmpbuf[0] == (unsigned char) 0x0D &&
	      cmpbuf[1] == (unsigned char) 0x0D) ||
	     (cmpbuf[0] == (unsigned char) 0x0A &&
	      cmpbuf[1] == (unsigned char) 0x0A)) {
      /* Replace with 0x0A 0x0A */
      buf[returnSZ] = 0x0A;
      returnSZ ++;
      buf[returnSZ] = 0x0A;
      returnSZ ++;
      cmpbuf[0] = cmpbuf[2];
      cmpbuf[1] = cmpbuf[3];
      cmpbufSZ = 2;
    }
    /* Is it stray linefeeds here? */
    else if ((cmpbuf[0] == (unsigned char) 0x0D) ||
	     (cmpbuf[0] == (unsigned char) 0x0A)) {
      cmpbuf[0] = cmpbuf[1];
      cmpbuf[1] = cmpbuf[2];
      cmpbuf[2] = cmpbuf[3];
      cmpbufSZ = 3;
    }
    /* Just push it out then */
    else {
      buf[returnSZ] = cmpbuf[0];
      returnSZ ++;
      cmpbuf[0] = cmpbuf[1];
      cmpbuf[1] = cmpbuf[2];
      cmpbuf[2] = cmpbuf[3];
      cmpbufSZ = 3;
    }
  }
  return (returnSZ);
}

/* MAIN PROGRAM */

int main(int argc, char *argv[])
{
  int arg, i;

  bool bReadRaw;
  bool bReadText;
  bool bCompress;
  bool bDecompress;
  bool bInfo;
  bool bHexDump;
  bool bStdout;

  char cInFile[STR_MAXLEN] = "";
  char cOutFile[STR_MAXLEN] = "";

  DWORD ChunkSize;
  static byte buf[MAGIC_CHUNKSIZE];

  FILE *fin, *fout;
  tPilotRecord *PilotRecords = NULL;

  if (argc<2) {
    printf("Usage: %s [-options] <infile> [<outfile>]\n", argv[0]);
    printf("Type %s -help for all options.\n", argv[0]);
    exit(0);
  }

  bReadText = false;
  bReadRaw = false;
  bCompress = false;
  bDecompress = false;
  bInfo = false;
  bHexDump = false;
  bStdout = false;

  for (arg=1; arg<argc; arg++) {
    if (*argv[arg] == '-') {
      /* -? or -help: show the command line help page */
      if (!strcmp(argv[arg],"-?") || !strcmp(argv[arg],"-help") || !strcmp(argv[arg],"--help")) {
	printf("Usage: %s [-options] <infile[.pdb]> [<outfile[.pdb]>]\n\n", argv[0]);
	printf("Options: -i      = Print PDB file info while working\n");
	printf("         -c      = Compress PDB file chunks (DOC style compression)\n");
	printf("         -d      = Decompress PDB file chunks (DOC style compression)\n");
	printf("         -h      = hex dump of PDB file\n");
	printf("         -raw    = Read a raw file (.txt or other) and create records\n");
	printf("         -txt    = Read a text file and create a PDB (MakeDoc mode)\n");
	printf("         -       = Write to stdout\n\n");
	printf("Example: create a PDB DOC file from a plain text file:\n");
	printf("  %s -i -c -txt plainfile.txt palmdoc.pdb\n", argv[0]);
	printf("Example: decompress a compressed DOC document and view it:\n");
	printf("  %s -d - palmdoc.pdb | more\n", argv[0]);
	exit(0);
	/* non-interactive */
      }
      if (!strcmp(argv[arg],"-raw"))
	bReadRaw = true;
      if (!strcmp(argv[arg],"-txt")) {
	bReadText = true;
	/* Text is implicitly raw */
	bReadRaw = true;
      }
      if (!strcmp(argv[arg],"-c"))
	bCompress = true;
      if (!strcmp(argv[arg],"-d"))
	bDecompress = true;
      if (!strcmp(argv[arg],"-i"))
	bInfo = true;
      if (!strcmp(argv[arg], "-h"))
	bHexDump = true;
      if (!strcmp(argv[arg],"-"))
	bStdout = true;
    }
  }

  /* Both input and output filename given */
  if (*argv[argc-1] != '-' && *argv[argc-2] != '-' && argc > 2) {
    strcpy(cInFile, argv[argc-2]);
    strcpy(cOutFile, argv[argc-1]);
  } else if (*argv[argc-1] != '-' || argc == 2) {
    strcpy(cInFile, argv[argc-1]);
    if (bInfo)
      fprintf(stderr, "Only input file name give, writing to stdout...\n");
    bStdout = true;
  } else {
    fprintf(stderr, "No input file, dork...\n");
    return (-1);
  }

  /* See that the source is OK */
  if (!bReadRaw) {
    if (strcmp(cInFile+(strlen(cInFile)-4), ".pdb"))
      strcpy(cInFile+strlen(cInFile), ".pdb");
  }
  if(bInfo)
    fprintf(stderr, "Input file: %s\n", cInFile);
  
  /* See that the destination is OK */
  if (!bStdout && !bHexDump) {
    if (strcmp(cOutFile+(strlen(cOutFile)-4), ".pdb"))
      strcpy(cOutFile+strlen(cOutFile), ".pdb");
    if(bInfo)
      fprintf(stderr, "Output file: %s\n", cOutFile);
  } else if (bHexDump) {
    if (bInfo)
      fprintf(stderr, "Creating hex dump on stdout from file...\n");
  }

  if((fin = fopen(cInFile, "rb"))) {
    if (bStdout)
      fout = stdout;
    else if(!(fout = fopen(cOutFile, "wb"))) {
      fprintf(stderr, "%s: failed to open file %s, dork ..\n", argv[0], cOutFile);
      fclose(fin);
      exit(-1);
    }
    
    /* Reading in raw data or PDB file */
    if(bReadRaw) {
      if (bInfo)
	fprintf(stderr, "Reading raw file...\n");
      SetPDBFileType("TEXt", "REAd", RECATT_DIRTY);
      i = 0;
      if (bReadText) {
	if (bInfo)
	  fprintf(stderr, "Stripping extraneous linefeeds in text file...\n");
	while((ChunkSize = txtfread(buf, MAGIC_CHUNKSIZE, fin))) {
     	  if (bInfo)
	    fprintf(stderr, "Adding chunk %d size %d bytes .. \n", i++, (int) ChunkSize);
	  PilotRecords = AddRecord(PilotRecords, buf, (DWORD) ChunkSize, (DWORD) 0);
	}
      } else {
	while((ChunkSize = fread(buf, 1, MAGIC_CHUNKSIZE, fin))) {
	  if (bInfo)
	    fprintf(stderr, "Adding chunk %d size %d bytes .. \n", i++, (int) ChunkSize);
	  PilotRecords = AddRecord(PilotRecords, buf, (DWORD) ChunkSize, (DWORD) 0);
	}
      }
    } else {
      /* Reading a PDB file */
      if (bInfo)
	fprintf(stderr, "Reading PDB file...\n");
      PilotRecords = ReadPDBFile(fin);
      /* To force this type */
      /* SetPDBFileType("TEXt", "REAd", RECATT_DIRTY); */
    }
    
    /* Handle compression / decompression */
    if (bCompress) {
      if (bInfo)
	fprintf(stderr, "Compressing chunks...\n");
      DocCompressList(PilotRecords);
    }
    if (bDecompress) {
      if (bInfo)
	fprintf(stderr, "Decompressing chunks...\n");
      DeDocCompressList(PilotRecords);
    }

    /* Dump or write the resulting file */
    if (bHexDump) {
      DumpRecords(PilotRecords, fout);
    } else {
      WritePDBFile(PilotRecords, fout, cOutFile);
      if (bInfo)
	fprintf(stderr, "PDB file \"%s\" created.\n", cOutFile);
    }
      
    /* Free memory used by database records */
    FreeRecords(PilotRecords);
    
    if (!bStdout)
      fclose(fout);

  } else
    fprintf(stderr, "%s: failed to open file %s, dork ...\n", argv[0], cInFile);
  return 0;
}
